﻿/* 
 * Copyright (C) 2008 Alan Ross, Tojio GmbH (http://lab.tojio.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
 

package
{	
	import flash.display.*;
	import flash.events.*;
	import flash.filters.*;
	import flash.geom.*;
	import flash.text.TextField;
	//import Papervision (effects branch)
	import org.papervision3d.cameras.*;
	import org.papervision3d.events.*;
	import org.papervision3d.core.effects.*;
	import org.papervision3d.core.layers.*;
	import org.papervision3d.lights.*;
	import org.papervision3d.materials.*;
	import org.papervision3d.materials.shaders.*;
	import org.papervision3d.materials.utils.*;
	import org.papervision3d.objects.*;
	import org.papervision3d.objects.primitives.*;
	import org.papervision3d.view.*;
	//import WiiFlash (0.4)
	import org.wiiflash.*;
	import org.wiiflash.events.*;

    //[SWF(frameRate="31", width="900", height="600", backgroundColor="0x111111")]
    public class WiiFlash3DHeadTracking extends Sprite
    {
    	//Papervision:
      	private const NUM_OF_OBJECTS:uint = 8;
      	//either use cubes or spheres
        private const USE_CUBES:Boolean = true;
        //change amount of x, y and z are to be moved
        private const FACTOR_XY:uint = 200;
        private const FACTOR_Z:uint = 10;
        //basic objects for 3d world
        private var view:BasicView;
        private var demoObjects:DisplayObject3D;
        private var environment:Cube;
        private var movement:Object;
		//used for mouse interaction
		private var activeObj:DisplayObject3D;
		private var activeObjDir:int = 0;
		
		//Wiimote:
		private var wiimote:Wiimote;
        
		//debugging:
		private var debug:TextField;
		
		//________________________________________________________
		// constructor
		//________________________________________________________
		public function WiiFlash3DHeadTracking()
		{   
            super();
			if(stage != null){ init(); }
			else{ addEventListener(Event.ADDED_TO_STAGE, init); }
		}
		
		
		//________________________________________________________
		// initialize
		//________________________________________________________
		public function init(e:Event=null):void 
		{	
			stage.scaleMode = StageScaleMode.NO_SCALE;
        	stage.align = StageAlign.TOP_LEFT;
        	stage.quality = StageQuality.HIGH;
        	
        	debug = new TextField(); debug.height = 400; debug.width = 300; 
        	debug.x = debug.y = 10; debug.textColor = 0xffffff; debug.selectable = false; 
        	addChild(debug);
        	
        	movement = new Object();
        	movement.newX = 0;
			movement.newY = 0;
			movement.newZ = 0;
			movement.oldX = 0;
			movement.oldY = 0;
			movement.oldZ = 0;
        	
        	initWiiFlash();
        	init3D();
        	
            addEventListener( Event.ENTER_FRAME, loop ); 
		}
		
		
		//________________________________________________________
		// setup papervision
		//________________________________________________________
		private function init3D():void
		{
			view = new BasicView(stage.stageWidth, stage.stageHeight, true, true);
        	view.camera.zoom = 4;
        	view.camera.focus = 130;
        	view.camera.z = -1000;
        	addChildAt(view,0);
	
			//the root node
			demoObjects = new DisplayObject3D();
			
			//the environment
			var m:WireframeMaterial = new WireframeMaterial(0x333333,.2);
			var ml:MaterialsList = new MaterialsList({front:m, back:m, right:m, left:m, top:m, bottom:m});
			var s:uint = 1000;	//size
			var q:uint = 10;	//quality
			environment = new Cube(ml,s,s,s,q,q,q,Cube.ALL);
			environment.z = -500;
			view.scene.addChild(environment, "environment");
			
			//light and shading (used if shaded material is applied)
			var light:PointLight3D = new PointLight3D(true, false);
			light.z = -1000;
			light.x = 500;
        	light.y = 500;
        	
        	//now add some demo objects to play around with
			var o:DisplayObject3D;
			var darkenRate:Number;
			var blurRate:Number;
			var ct:ColorTransform;
			var bmd:BitmapData;
			var bm:BitmapMaterial;
			var sm:ShadedMaterial;
			var fx:EffectLayer;
			var cs:Shader = new GouraudShader(light, 0xFFFFFF,0x555555);
        	
			for(var i:uint=0; i<NUM_OF_OBJECTS; i++)
			{
				darkenRate = 1-i/(NUM_OF_OBJECTS*1.3);
				blurRate = i/2;
				ct = new ColorTransform(darkenRate,darkenRate,darkenRate,1);
				bmd = new BitmapData(500,500);
				bmd.perlinNoise(10,10,3,Math.random()*10,true,true);
				bmd.colorTransform(bmd.rect,ct);
				bm = new BitmapMaterial(bmd);
				sm = new ShadedMaterial(bm,cs);
			
				fx = new EffectLayer();
				view.viewport.addRenderLayer(fx);
				fx.addEffect(new LayerEffect(new BlurFilter(blurRate,blurRate,1)));
				
				if(USE_CUBES)
					o = new Cube(new MaterialsList({all:bm}),100,10,100,1,1,1);
				else
					o = new Sphere(sm,40,12,8);
				
				o.renderLayer = fx;
				o.y = Math.random()*400-200;
				o.x = Math.random()*400-200;
				o.z = (i*100)-800;
				o.yaw(Math.random()*180);
        		o.roll(Math.random()*180);
        		o.pitch(Math.random()*180);
				o.material.interactive = true;
				o.addEventListener(InteractiveScene3DEvent.OBJECT_PRESS, onDown);
				o.addEventListener(InteractiveScene3DEvent.OBJECT_RELEASE, onUp);
				o.addEventListener(InteractiveScene3DEvent.OBJECT_RELEASE_OUTSIDE, onUp);
				demoObjects.addChild(o);
			}
			
			//add blur to environment too
			environment.renderLayer = fx;
			//now add demo objects to scene 
        	view.scene.addChild(demoObjects);
		}
		
		//________________________________________________________
		// setup papervision
		//________________________________________________________
		private function initWiiFlash():void
		{
			wiimote = new Wiimote();
			wiimote.connect();
			
			wiimote.addEventListener( Event.CONNECT, onWiimoteConnect );
			wiimote.addEventListener( IOErrorEvent.IO_ERROR, onWiimoteConnectError );
			wiimote.addEventListener( Event.CLOSE, onCloseConnection );
			wiimote.addEventListener( WiimoteEvent.UPDATE, onUpdated );
		}
		
		
		//________________________________________________________
		// main rendering loop
		//________________________________________________________
        private function loop( event : Event ):void
        {
			// rotate selected object
        	if(activeObj != null)
        	{
        		activeObj.yaw(activeObjDir);
        		activeObj.roll(activeObjDir);
        		activeObj.pitch(activeObjDir);	
        	}
        	//if wii ir dots 1 and 2 are active
        	if(wiimote.ir.p1 && wiimote.ir.p2)
			{
				movement.newX = ((wiimote.ir.x1+wiimote.ir.x2)-1)/2;
				movement.newY = ((wiimote.ir.y1+wiimote.ir.y2)+1)/2;
				movement.newZ = Math.abs(wiimote.ir.x1-wiimote.ir.x2)*100;
				
				movement.newX *= -FACTOR_XY;
				movement.newY *= -FACTOR_XY;
				movement.newZ *= FACTOR_Z;
				
				movement.newX = (int)((movement.newX+movement.oldX)/2);
				movement.newY = (int)((movement.newY+movement.oldY)/2);
				movement.newZ = (int)((movement.newZ+movement.oldZ)/2);
				
				movement.oldX = movement.newX;
				movement.oldY = movement.newY;
				movement.oldZ = movement.newZ;
				
				Camera3D(view.camera).x = movement.newX;
				Camera3D(view.camera).y = movement.newY;
				Camera3D(view.camera).z = -1100 - (movement.newZ/2);
				//let objects 'pop out' a little bit...
				demoObjects.z = -movement.newZ;
			}
			
			view.singleRender();
        }


		

		private function onDown( e:Event ):void
		{ activeObj = (DisplayObject3D)(e.target); activeObjDir = 20; }
		
		private function onUp( e:Event ):void
		{ activeObj = null; }
	

		private function onCloseConnection( e:Event ):void
		{ debug.text = "Connection closed, WiiFlash Server has been closed"; }
		
		private function onWiimoteConnectError( e:IOErrorEvent ):void
		{ debug.text = "Couldn't connect, make sure WiiFlash Server is running !"; }
	
		private function onWiimoteConnect( e:Event ):void
		{ debug.text= "Wiimote successfully connected"; }
	
		private function onUpdated( e:WiimoteEvent ):void 
		{
			debug.text = "Wiimote: " + wiimote.id + "\n";
			
			debug.appendText("IR x1 : " + String ( wiimote.ir.x1 ).substr(0,6) + "\n");
			debug.appendText("IR y1 : " + String ( wiimote.ir.y1 ).substr(0,6) + "\n");
			debug.appendText("IR x2 : " + String ( wiimote.ir.x2 ).substr(0,6) + "\n");
			debug.appendText("IR y2 : " + String ( wiimote.ir.y2 ).substr(0,6) + "\n");
			debug.appendText("IR x3 : " + String ( wiimote.ir.x3 ).substr(0,6) + "\n");
			debug.appendText("IR y3 : " + String ( wiimote.ir.y3 ).substr(0,6) + "\n");
			debug.appendText("IR x4 : " + String ( wiimote.ir.x4 ).substr(0,6) + "\n");
			debug.appendText("IR y4 : " + String ( wiimote.ir.y4 ).substr(0,6) + "\n");
		}
    }
}